import numpy


def rotation_matrix(u, theta):
    """Return matrix that implements the rotation around the vector :math:`u`
    by the angle :math:`\\theta`, cf.
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle.

    :param u: rotation vector
    :param theta: rotation angle
    """
    assert numpy.isclose(numpy.inner(u, u), 1.0), "the rotation axis must be unitary"

    # Cross-product matrix.
    cpm = numpy.array([[0.0, -u[2], u[1]], [u[2], 0.0, -u[0]], [-u[1], u[0], 0.0]])
    c = numpy.cos(theta)
    s = numpy.sin(theta)
    R = numpy.eye(3) * c + s * cpm + (1.0 - c) * numpy.outer(u, u)
    return R


def orient_lines(lines):
    """Given a sequence of unordered and unoriented lines defining a closed polygon,
    returns a reordered list of reoriented lines of that polygon.

    :param lines: a sequence of lines defining a closed polygon
    """
    # Categorise graph edges by their vertex pair ids
    point_pair_ids = numpy.array(
        [[line.points[0]._ID, line.points[1]._ID] for line in lines]
    )

    # Indices of reordering
    order = numpy.arange(len(point_pair_ids), dtype=int)
    # Compute orientations where oriented[j] == False requires edge j to be reversed
    oriented = numpy.array([True] * len(point_pair_ids), dtype=numpy.bool)

    for j in range(1, len(point_pair_ids)):
        out = point_pair_ids[j - 1, 1]  # edge out from vertex
        inn = point_pair_ids[j:, 0]  # candidates for edge into vertices
        wh = numpy.where(inn == out)[0] + j
        if len(wh) == 0:
            # look for candidates in those which are not correctly oriented
            inn = point_pair_ids[j:, 1]
            wh = numpy.where(inn == out)[0] + j
            # reorient remaining edges
            point_pair_ids[j:] = numpy.flip(point_pair_ids[j:], axis=1)
            oriented[j:] ^= True

        # reorder
        point_pair_ids[[j, wh[0]]] = point_pair_ids[[wh[0], j]]
        order[[j, wh[0]]] = order[[wh[0], j]]

    # Reconstruct an ordered and oriented line loop
    lines = [lines[o] for o in order]
    lines = [lines[j] if oriented[j] else -lines[j] for j in range(len(oriented))]

    return lines



import argparse
from sys import version_info

import gmsh
import meshplex
import meshio
import numpy
import termplotlib
from .__about__ import __version__, __gmsh_version__


def optimize(argv=None):
    parser = argparse.ArgumentParser(
        description=("Optimize mesh."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("infile", type=str, help="mesh to optimize")
    parser.add_argument("outfile", type=str, help="optimized mesh")

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=_get_version_text(),
        help="display version information",
    )
    args = parser.parse_args(argv)

    mesh = meshio.read(args.infile)
    mesh.remove_lower_dimensional_cells()
    mesh.cell_data = {}

    meshplex_mesh = meshplex.from_meshio(mesh)
    print_stats(meshplex_mesh)

    # optimize
    tmp_filename = "tmp.msh"
    mesh.write(tmp_filename)
    gmsh.initialize()
    gmsh.model.add("Mesh from file")
    gmsh.merge(tmp_filename)
    gmsh.model.mesh.optimize("Netgen")
    gmsh.write(tmp_filename)
    gmsh.finalize()
    mesh = meshio.read(tmp_filename)
    mesh.write(args.outfile)
    return


def _get_version_text():
    return "\n".join(
        [
            f"pygmsh {__version__} "
            f"[Gmsh {__gmsh_version__}, "
            f"Python {version_info.major}.{version_info.minor}.{version_info.micro}]",
            "Copyright (c) 2013-2020 Nico Schlömer et al.",
        ]
    )


def print_stats(mesh, extra_cols=None):
    extra_cols = [] if extra_cols is None else extra_cols

    # angles = mesh.angles / numpy.pi * 180
    # angles_hist, angles_bin_edges = numpy.histogram(
    #     angles, bins=numpy.linspace(0.0, 180.0, num=73, endpoint=True)
    # )

    q = mesh.q_radius_ratio
    q_hist, q_bin_edges = numpy.histogram(
        q, bins=numpy.linspace(0.0, 1.0, num=41, endpoint=True)
    )
    print(q_hist)

    grid = termplotlib.subplot_grid(
        (1, 4 + len(extra_cols)), column_widths=None, border_style=None
    )
    # grid[0, 0].hist(angles_hist, angles_bin_edges, grid=[24], bar_width=1, strip=True)
    # grid[0, 1].aprint("min angle:     {:7.3f}".format(numpy.min(angles)))
    # grid[0, 1].aprint("avg angle:     {:7.3f}".format(60))
    # grid[0, 1].aprint("max angle:     {:7.3f}".format(numpy.max(angles)))
    # grid[0, 1].aprint("std dev angle: {:7.3f}".format(numpy.std(angles)))
    grid[0, 2].hist(q_hist, q_bin_edges, bar_width=1, strip=True)
    grid[0, 3].aprint("min quality: {:5.3f}".format(numpy.min(q)))
    grid[0, 3].aprint("avg quality: {:5.3f}".format(numpy.average(q)))
    grid[0, 3].aprint("max quality: {:5.3f}".format(numpy.max(q)))

    for k, col in enumerate(extra_cols):
        grid[0, 4 + k].aprint(col)

    grid.show()
